package net.sergeych.kiloparsec.adapter

import assertThrows
import kotlinx.coroutines.test.runTest
import net.sergeych.crypto2.initCrypto
import net.sergeych.kiloparsec.*
import net.sergeych.mp_logger.Log
import kotlin.random.Random
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertIs

class InternetTest {
    class TestException : Exception("test1")

    @Test
    fun tcpTest() = runTest {
        initCrypto()
        Log.connectConsole(Log.Level.DEBUG)
        data class Session(
            var data: String,
        )

        val port = 27170 + Random.nextInt(1, 200)

        val cmdSave by command<String, Unit>()
        val cmdLoad by command<Unit, String>()
        val cmdDrop by command<Unit, Unit>()
        val cmdPing by command<String, String>()
        val cmdException by command<Unit, Unit>()
        val cmdCallClient by command<String, String>()

        val serverInterface = KiloInterface<Session>().apply {
            registerError { TestException() }
            onConnected { session.data = "start" }
            on(cmdSave) { session.data = it }
            on(cmdLoad) {
                session.data
            }
            on(cmdException) {
                throw TestException()
            }
            on(cmdDrop) {
                throw LocalInterface.BreakConnectionException()
            }
            on(cmdCallClient) {
                remote.call(cmdPing, it)
            }
        }
        val server = KiloServer(serverInterface, acceptTcpDevice(port)) {
            Session("unknown")
        }

        data class LocalSession(val localFoo: String)

        val client = KiloClient {
            addErrors(serverInterface)
            session { LocalSession("unknown") }
            // TODO: add register error variant
            connect { connectTcpDevice("localhost:$port") }
            local {
                on(cmdPing) {
                    "pong! $it"
                }
            }
        }

        assertEquals("start", client.call(cmdLoad))

        client.call(cmdSave, "foobar")
        assertEquals("foobar", client.call(cmdLoad))

        val res = kotlin.runCatching { client.call(cmdException) }
        assertIs<TestException>(res.exceptionOrNull())
        assertEquals("foobar", client.call(cmdLoad))

        assertThrows<RemoteInterface.ClosedException> { client.call(cmdDrop) }

        // reconnect?
        assertEquals("start", client.call(cmdLoad))
        assertEquals("pong! 42", client.call(cmdCallClient, "42"))
        server.close()
    }

    @Test
    fun udpTest() = runTest {
        initCrypto()
        Log.connectConsole(Log.Level.DEBUG)
        data class Session(
            var data: String,
        )

        val port = 27170 + Random.nextInt(1, 200)

        val cmdSave by command<String, Unit>()
        val cmdLoad by command<Unit, String>()
        val cmdDrop by command<Unit, Unit>()
        val cmdException by command<Unit, Unit>()

        val cli = KiloInterface<Session>().apply {
            registerError { TestException() }
            onConnected { session.data = "start" }
            on(cmdSave) { session.data = it }
            on(cmdLoad) {
                session.data
            }
            on(cmdException) {
                throw TestException()
            }
            on(cmdDrop) {
                throw LocalInterface.BreakConnectionException()
            }
        }
        val server = KiloServer(cli, acceptUdpDevice(port)) {
            Session("unknown")
        }

        val client = KiloClient<Unit>() {
            addErrors(cli)
            connect { connectUdpDevice("127.0.0.1:$port") }
        }

        assertEquals("start", client.call(cmdLoad))

        client.call(cmdSave, "foobar")
        assertEquals("foobar", client.call(cmdLoad))

        val res = kotlin.runCatching { client.call(cmdException) }
        assertIs<TestException>(res.exceptionOrNull())
        assertEquals("foobar", client.call(cmdLoad))

        assertThrows<RemoteInterface.ClosedException> { client.call(cmdDrop) }

        // reconnect?
        assertEquals("start", client.call(cmdLoad))

        server.close()
    }

    @Test
    fun udpServerTest() = runTest {
        initCrypto()
        Log.connectConsole(Log.Level.DEBUG)
        data class Session(
            var data: String,
        )

        val port = 27170 + Random.nextInt(1, 900)

        val cmdSave by command<String, Unit>()
        val cmdLoad by command<Unit, String>()
        val cmdDrop by command<Unit, Unit>()
        val cmdException by command<Unit, Unit>()

        val cli = KiloInterface<Session>().apply {
            registerError { TestException() }
            onConnected { session.data = "start" }
            on(cmdSave) { session.data = it }
            on(cmdLoad) {
                session.data
            }
            on(cmdException) {
                throw TestException()
            }
            on(cmdDrop) {
                throw LocalInterface.BreakConnectionException()
            }
        }
        val uServer = UdpServer(port)

        KiloServer(cli, uServer.transportFlow) {
            Session("unknown")
        }

        // second server
        KiloServer(cli,acceptUdpDevice(uServer.port+1)) { Session("unknown2") }


        val client = KiloClient<Unit>() {
            addErrors(cli)
            connect { connectUdpDevice("127.0.0.1:$port") }
        }
        val client2 = KiloClient<Unit>() { connect { connectUdpDevice("127.0.0.1:${port+1}") } }

        assertEquals("start", client.call(cmdLoad))
        assertEquals("start", client2.call(cmdLoad))

        client.call(cmdSave, "foobar")
        client2.call(cmdSave, "buzz")
        assertEquals("foobar", client.call(cmdLoad))
        assertEquals("buzz", client2.call(cmdLoad))

        val res = kotlin.runCatching { client.call(cmdException) }
        assertIs<TestException>(res.exceptionOrNull())
        assertEquals("foobar", client.call(cmdLoad))

        assertThrows<RemoteInterface.ClosedException> { client.call(cmdDrop) }

        // reconnect?
        assertEquals("start", client.call(cmdLoad))

        uServer.close()
//        server.close()
    }
}